import platform
from pprint import pprint
import sys
import os
from pathlib import Path
from string import Template
import re
import shutil
import textwrap
from typing import Any, Dict, List, Optional, Union
import tempfile
from subprocess import CalledProcessError, run as sp_run
from glob import glob

from .config import Config


class _BuildBackend(object):

    def __init__(self) -> None:
        self.verbose = False

    def check_program_version(
        self,
        program: str,
        minimum_version: Optional[str],
        name: Optional[str],
        check_version: bool = True,
    ):
        """Check if there's a new enough version of the given command available
        in PATH."""
        from distlib.version import NormalizedVersion
        name = name or program
        try:
            # Try running the command
            cmd = [program, '--version'] if check_version else [program, '-h']
            res = self.run(cmd, check=True, capture_output=True)
            # Try finding the version
            if check_version:
                m = re.search(r'\d+(\.\d+){1,}', res.stdout.decode('utf-8'))
                if not m:
                    raise RuntimeError(f"Unexpected {name} version output")
                program_version = NormalizedVersion(m.group(0))
                if self.verbose: print("Found", name, program_version)
                # Check if the version is new enough
                if minimum_version:
                    if program_version < NormalizedVersion(minimum_version):
                        raise RuntimeError(f"{name} too old")
        except CalledProcessError as e:
            if self.verbose:
                print(f'{type(e).__module__}.{type(e).__name__}', e, sep=': ')
                sys.stdout.buffer.write(e.stdout)
                sys.stdout.buffer.write(e.stderr)
            return False
        except Exception as e:
            # If any of that failed, return False
            if self.verbose:
                print(f'{type(e).__module__}.{type(e).__name__}', e, sep=': ')
            return False
        return True

    def get_requires_for_build_wheel(self, config_settings=None):
        """https://www.python.org/dev/peps/pep-0517/#get-requires-for-build-wheel"""
        self.parse_config_settings(config_settings)
        pyproject = Path('pyproject.toml').resolve()
        cfg = self.read_metadata(pyproject)
        deps = []
        # Check if we need CMake
        if cfg.cmake:
            self.check_cmake_program(cfg, deps)
        if cfg.stubgen:
            self.check_stubgen_program(cfg, deps)
        if self.verbose:
            print('Dependencies for build:', deps)
        return deps

    def check_cmake_program(self, cfg: Config, deps: List[str]):
        assert cfg.cmake
        min_cmake_ver = cfg.cmake[self.get_os_name()].get('minimum_version')
        # If CMake in PATH doesn't work, add it as a build requirement
        if not self.check_program_version('cmake', min_cmake_ver, "CMake"):
            req = "cmake"
            if min_cmake_ver: req += ">=" + min_cmake_ver
            deps.append(req)
        # Check if we need Ninja
        cfgs = []
        # Native build?
        native = not cfg.cross
        crossnative = self.needs_cross_native_build(cfg)
        if native or crossnative:
            cfgs.append(cfg.cmake[self.get_os_name()])
        # Cross build?
        cross = cfg.cross
        if cross:
            cfgs.append(cfg.cmake['cross'])
        # Do any of the configs need Ninja as a generator?
        needs_ninja = lambda c: 'ninja' in c.get('generator', '').lower()
        need_ninja = any(map(needs_ninja, cfgs))
        if need_ninja:
            if not self.check_program_version('ninja', None, "Ninja"):
                deps.append("ninja")

    def check_stubgen_program(self, cfg: Config, deps: List[str]):
        if not self.check_program_version('stubgen', None, None, False):
            deps.append("mypy")

    def get_requires_for_build_editable(self, config_settings=None):
        """https://www.python.org/dev/peps/pep-0660/#get-requires-for-build-editable"""
        return self.get_requires_for_build_wheel(config_settings)

    def get_requires_for_build_sdist(self, config_settings=None):
        """https://www.python.org/dev/peps/pep-0517/#get-requires-for-build-sdist"""
        return []

    def build_wheel(self,
                    wheel_directory,
                    config_settings=None,
                    metadata_directory=None):
        """https://www.python.org/dev/peps/pep-0517/#build-wheel"""
        assert metadata_directory is None

        # Parse options
        self.parse_config_settings(config_settings)

        # Build wheel
        with tempfile.TemporaryDirectory() as tmp_build_dir:
            whl_name = self.build_wheel_in_dir(wheel_directory, tmp_build_dir)
        return whl_name

    def build_editable(self,
                       wheel_directory,
                       config_settings=None,
                       metadata_directory=None):
        """https://www.python.org/dev/peps/pep-0660/#build-editable"""
        assert metadata_directory is None

        # Parse options
        self.parse_config_settings(config_settings)

        # Build wheel
        with tempfile.TemporaryDirectory() as tmp_build_dir:
            whl_name = self.build_wheel_in_dir(wheel_directory,
                                               tmp_build_dir,
                                               editable=True)
        return whl_name

    def build_sdist(self, sdist_directory, config_settings=None):
        """https://www.python.org/dev/peps/pep-0517/#build-sdist"""
        sdist_directory = Path(sdist_directory)
        src_dir = Path().resolve()

        # Parse options
        self.parse_config_settings(config_settings)

        # Load metadata
        from .config import normalize_name_wheel
        from flit_core.common import Module, make_metadata
        pyproject = src_dir / 'pyproject.toml'
        cfg = self.read_metadata(pyproject)
        import_name = cfg.module['name']
        pkg = Module(import_name, src_dir / cfg.module['directory'])
        metadata = make_metadata(pkg, cfg)
        metadata.version = self.normalize_version(metadata.version)

        # Export dist
        from flit_core.sdist import SdistBuilder
        rel_pyproject = os.path.relpath(pyproject, src_dir)
        extra_files = [str(rel_pyproject)] + cfg.referenced_files
        sdist_cfg = cfg.sdist[self.get_os_name()]
        sdist_builder = SdistBuilder(
            pkg,
            metadata=metadata,
            cfgdir=src_dir,
            reqs_by_extra=None,
            entrypoints=cfg.entrypoints,
            extra_files=extra_files,
            data_directory=None,
            include_patterns=sdist_cfg.get('include_patterns', []),
            exclude_patterns=sdist_cfg.get('exclude_patterns', []),
        )
        sdist_tar = sdist_builder.build(Path(sdist_directory))
        return os.path.relpath(sdist_tar, sdist_directory)

    @staticmethod
    def get_os_name():
        return {
            "Linux": "linux",
            "Windows": "windows",
            "Darwin": "mac",  # TODO: untested
        }[platform.system()]

    def parse_config_settings(self, config_settings: Optional[Dict]):
        if 'PY_BUILD_CMAKE_VERBOSE' in os.environ:
            self.verbose = True
        if config_settings is None:
            return
        if config_settings.keys() & {'verbose', 'v'}:
            self.verbose = True

    def read_metadata(self, pyproject):
        from .config import read_metadata
        cfg = read_metadata(pyproject)
        if self.verbose:
            print("\npy-build-cmake options")
            print("======================")
            print("module:")
            pprint(cfg.module)
            print("sdist:")
            pprint(cfg.sdist)
            print("cmake:")
            pprint(cfg.cmake)
            print("stubgen:")
            pprint(cfg.stubgen)
            print("cross:")
            pprint(cfg.cross)
            print("======================\n")
        return cfg

    def build_wheel_in_dir(self,
                           wheel_directory,
                           tmp_build_dir,
                           editable=False):
        wheel_directory = Path(wheel_directory)
        tmp_build_dir = Path(tmp_build_dir)
        staging_dir = tmp_build_dir / 'staging'
        src_dir = Path().resolve()

        # Load metadata
        from .config import normalize_name_wheel
        from flit_core.common import Module, make_metadata
        cfg = self.read_metadata(src_dir / 'pyproject.toml')
        norm_name = normalize_name_wheel(cfg.metadata['name'])
        import_name = cfg.module['name']
        pkg = Module(import_name, src_dir / cfg.module['directory'])
        metadata = make_metadata(pkg, cfg)
        metadata.version = self.normalize_version(metadata.version)

        # Copy the module source files to the temporary folder
        if not editable:
            self.copy_pkg_source_to(staging_dir, src_dir, pkg)
        else:
            self.write_editable_wrapper(staging_dir, src_dir, pkg)

        # Create dist-info folder
        distinfo = staging_dir / f'{norm_name}-{metadata.version}.dist-info'
        os.makedirs(distinfo, exist_ok=True)

        # Write metadata
        metadata_path = distinfo / 'METADATA'
        with open(metadata_path, 'w', encoding='utf-8') as f:
            metadata.write_metadata_file(f)

        # Write or copy license
        self.write_license_files(cfg.license, src_dir, distinfo)

        # Write entrypoints
        self.write_entrypoints(distinfo, cfg)

        # Generate .pyi stubs
        if cfg.stubgen is not None and not editable:
            self.generate_stubs(tmp_build_dir, staging_dir, pkg, cfg.stubgen)

        # Configure, build and install the CMake project
        if cfg.cmake:
            self.do_native_cross_cmake_build(tmp_build_dir, staging_dir,
                                             src_dir, cfg, metadata, norm_name)

        # Create wheel
        whl_name = self.create_wheel(wheel_directory, staging_dir, cfg,
                                     norm_name, metadata.version)
        return whl_name

    def needs_cross_native_build(self, cfg):
        return cfg.cross and 'copy_from_native_build' in cfg.cross

    def do_native_cross_cmake_build(self, tmp_build_dir, staging_dir, src_dir,
                                    cfg, metadata, norm_name):
        """If not cross-compiling, just do a regular CMake build+install.
        When cross-compiling, do a cross-build+install (using the provided 
        CMake toolchain file).
        If cfg.cross['copy_from_native_build'] is set, before cross-compiling, 
        first a normal build+install is performed to a separate directory, then
        the cross-build+install is performed, and finally the installed files
        from the native build that match the patterns in
        cfg.cross['copy_from_native_build'] are copied to the installation
        directory of the cross-build for packaging."""
        # When cross-compiling, optionally do a native build first
        native_install_dir = None
        native_cmake_cfg = cfg.cmake[self.get_os_name()]
        if self.needs_cross_native_build(cfg):
            native_install_dir = tmp_build_dir / 'native-install'
            self.run_cmake(src_dir, native_install_dir, metadata,
                           native_cmake_cfg, None, native_install_dir,
                           norm_name)
        # Then do the actual build
        cmake_cfg = cfg.cmake['cross'] if cfg.cross else native_cmake_cfg
        self.run_cmake(src_dir, staging_dir, metadata, cmake_cfg, cfg.cross,
                       native_install_dir, norm_name)
        # Finally, move the files from the native build to the staging area
        if native_install_dir:
            self.copy_native_install(staging_dir, native_install_dir,
                                     cfg.cross['copy_from_native_build'])

    def copy_native_install(self, staging_dir, native_install_dir,
                            native_install_patterns):
        """Copy the files that match the patterns from the native installation
        directory to the wheel staging directory."""
        for pattern in native_install_patterns:
            matches = sorted(glob(str(native_install_dir / pattern)))
            for path in matches:
                path = Path(path)
                rel = path.relative_to(native_install_dir)
                path.parent.mkdir(parents=True, exist_ok=True)
                print('-- Moving:', path, '->', staging_dir / rel.parent)
                shutil.move(path, staging_dir / rel.parent)
                # TODO: what if the folder already exists?
            if not matches:
                raise RuntimeError(
                    "Native build installed no files that matched the "
                    "pattern '" + pattern + "'")
        shutil.rmtree(native_install_dir)

    def normalize_version(self, version):
        from distlib.version import NormalizedVersion
        norm_version = str(NormalizedVersion(version))
        return norm_version

    def write_license_files(self, license, srcdir: Path, distinfo_dir: Path):
        if 'text' in license:
            with (distinfo_dir / 'LICENSE').open('w', encoding='utf-8') as f:
                f.write(license['text'])
        elif 'file' in license:
            shutil.copy2(srcdir / license['file'], distinfo_dir)

    def write_editable_wrapper(self, tmp_build_dir: Path, src_dir: Path, pkg):
        # Write a fake __init__.py file that points to the development folder
        tmp_pkg: Path = tmp_build_dir / pkg.name
        os.makedirs(tmp_pkg, exist_ok=True)
        special_dunders = [
            '__builtins__', '__cached__', '__file__', '__loader__', '__name__',
            '__package__', '__path__', '__spec__'
        ]
        content = f"""\
            # First extend the search path with the development folder
            __spec__.submodule_search_locations.insert(0, "{pkg.path}")
            # Now manually import the development __init__.py
            from importlib import util as _util
            _spec = _util.spec_from_file_location("{pkg.name}",
                                                  "{pkg.path}/__init__.py")
            _mod = _util.module_from_spec(_spec)
            _spec.loader.exec_module(_mod)
            # After importing, add its symbols to our global scope
            _vars = _mod.__dict__.copy()
            for _k in ['{"','".join(special_dunders)}']: _vars.pop(_k)
            globals().update(_vars)
            del _k
            del _spec
            del _mod
            del _vars
            del _util
            """
        (tmp_pkg / '__init__.py').write_text(textwrap.dedent(content),
                                             encoding='utf-8')
        # Add the py.typed file if it exists, so mypy picks up the stubs for
        # the C++ extensions
        py_typed: Path = pkg.path / 'py.typed'
        if py_typed.exists():
            shutil.copy2(py_typed, tmp_pkg)
        # Write a path file so IDEs find the correct files as well
        (tmp_build_dir / f'{pkg.name}.pth').write_text(str(pkg.path.parent))

    def copy_pkg_source_to(self, tmp_build_dir, src_dir, pkg):
        for mod_file in pkg.iter_files():
            rel_path = os.path.relpath(mod_file, pkg.path.parent)
            dst = tmp_build_dir / rel_path
            os.makedirs(dst.parent, exist_ok=True)
            shutil.copy2(mod_file, dst, follow_symlinks=False)

    def run(self, *args, **kwargs):
        """Wrapper around subprocess.run that optionally prints the command."""
        if self.verbose:
            pprint([*args])
            pprint(kwargs)
        return sp_run(*args, **kwargs)

    def generate_stubs(self, tmp_build_dir, staging_dir, pkg, cfg: Dict[str,
                                                                        Any]):
        """Generate stubs (.pyi) using mypy stubgen."""
        args = ['stubgen'] + cfg.get('args', [])
        cfg.setdefault('packages', [pkg.name] if pkg.is_package else [])
        for p in cfg['packages']:
            args += ['-p', p]
        cfg.setdefault('modules', [pkg.name] if not pkg.is_package else [])
        for m in cfg['modules']:
            args += ['-m', m]
        args += cfg.get('files', [])
        # Add output folder argument if not already specified in cfg['args']
        if 'args' not in cfg or not ({'-o', '--output'} & set(cfg['args'])):
            args += ['-o', staging_dir]
        env = os.environ.copy()
        env.setdefault('MYPY_CACHE_DIR', str(tmp_build_dir))
        # Call mypy stubgen in a subprocess
        self.run(args, cwd=pkg.path.parent, check=True, env=env)

    def run_cmake(self, pkgdir, install_dir, metadata, cmake_cfg, cross_cfg,
                  native_install_dir, norm_name):
        """Configure, build and install using CMake."""
        # Source and build folders
        srcdir = Path(cmake_cfg.get('source_path', pkgdir)).resolve()
        builddir = pkgdir / '.py-build-cmake_cache'
        builddir = Path(cmake_cfg.get('build_path', builddir))
        if not builddir.is_absolute():
            builddir = pkgdir / builddir
        buildconfig = self.get_build_config_name(cross_cfg)
        builddir = builddir / buildconfig
        builddir = builddir.resolve()
        # Environment variables
        cmake_env = os.environ.copy()
        if (f := 'env') in cmake_cfg:
            for k, v in cmake_cfg[f].items():
                cmake_env[k] = Template(v).substitute(cmake_env)

        # Configure
        configure_cmd = [
            'cmake',
            '-B',
            str(builddir),
            '-S',
            str(srcdir),
            '-D',
            'PY_BUILD_CMAKE_PACKAGE_VERSION:STRING=' + metadata.version,
            '-D',
            'PY_BUILD_CMAKE_PACKAGE_NAME:STRING=' + norm_name,
        ]
        if cross_cfg:
            toolchain = (pkgdir / cross_cfg['toolchain_file']).resolve()
            configure_cmd += [
                '-D', 'CMAKE_TOOLCHAIN_FILE:FILEPATH=' + str(toolchain), '-D',
                'Python3_EXECUTABLE:FILEPATH=' + sys.executable
            ]
        else:
            configure_cmd += [
                '-D', 'Python3_EXECUTABLE:FILEPATH=' + sys.executable, '-D',
                'Python3_ROOT_DIR:PATH=' + sys.prefix, '-D',
                'Python3_FIND_REGISTRY=NEVER', '-D',
                'Python3_FIND_STRATEGY=LOCATION'
            ]
        if native_install_dir:
            configure_cmd += [
                '-D', 'PY_BUILD_CMAKE_NATIVE_INSTALL_DIR:PATH=' +
                str(native_install_dir)
            ]
        configure_cmd += cmake_cfg.get('args', [])  # User-supplied arguments
        for k, v in cmake_cfg.get('options', {}).items():  # -D {option}={val}
            configure_cmd += ['-D', k + '=' + v]
        if btype := cmake_cfg.get('build_type'):  # -D CMAKE_BUILD_TYPE={type}
            configure_cmd += ['-D', 'CMAKE_BUILD_TYPE=' + btype]
        if gen := cmake_cfg.get('generator'):  # -G {generator}
            configure_cmd += ['-G', gen]
        self.run(configure_cmd, check=True, env=cmake_env)

        # Build and install
        if not cmake_cfg.get('config', False):
            self.build_install_cmake(cmake_cfg, builddir, cmake_env,
                                     install_dir, None)
        else:
            for config in cmake_cfg['config']:
                self.build_install_cmake(cmake_cfg, builddir, cmake_env,
                                         install_dir, config)

    def build_install_cmake(self, cmake_cfg, builddir, cmake_env, install_dir,
                            config):
        self.build_cmake(cmake_cfg, builddir, cmake_env, config)
        self.install_cmake(cmake_cfg, builddir, cmake_env, install_dir, config)

    def install_cmake(self, cmake_cfg, builddir, cmake_env, install_dir,
                      config):
        for component in cmake_cfg['install_components']:
            install_cmd = [
                'cmake', '--install',
                str(builddir), '--prefix',
                str(install_dir)
            ]
            install_cmd += cmake_cfg.get('install_args', [])
            if config:  # --config {config}
                install_cmd += ['--config', config]
            if component:
                install_cmd += ['--component', component]
            print('Installing component', component or 'all')
            self.run(install_cmd, check=True, env=cmake_env)

    def build_cmake(self, cmake_cfg, builddir, cmake_env, config):
        build_cmd = ['cmake', '--build', str(builddir)]
        build_cmd += cmake_cfg.get('build_args', [])  # User-supplied arguments
        if config:  # --config {config}
            build_cmd += ['--config', config]
        if (f := 'build_tool_args') in cmake_cfg:
            build_cmd += ['--'] + cmake_cfg[f]
        self.run(build_cmd, check=True, env=cmake_env)

    @staticmethod
    def get_build_config_name(cross_cfg):
        from distlib.wheel import IMPVER, ABI, ARCH
        buildconfig = '-'.join([IMPVER, ABI, ARCH])
        if cross_cfg:
            buildconfig = '-'.join(
                map(lambda x: x[0],
                    _BuildBackend.get_cross_tags(cross_cfg).values()))
        return buildconfig

    def write_entrypoints(self, distinfo: Path, cfg: Config):
        from flit_core.common import write_entry_points
        with (distinfo / 'entry_points.txt').open('w', encoding='utf-8') as f:
            write_entry_points(cfg.entrypoints, f)

    def create_wheel(self, wheel_directory, tmp_build_dir, cfg, norm_name,
                     norm_version):
        from distlib.wheel import Wheel
        whl = Wheel()
        whl.name = norm_name
        whl.version = norm_version
        libdir = 'platlib' if cfg.cmake else 'purelib'
        paths = {'prefix': str(tmp_build_dir), libdir: str(tmp_build_dir)}
        whl.dirname = wheel_directory
        tags = None
        if cfg.cross:
            tags = self.get_cross_tags(cfg.cross)
        wheel_path = whl.build(paths, tags=tags, wheel_version=(1, 0))
        whl_name = os.path.relpath(wheel_path, wheel_directory)
        return whl_name

    @staticmethod
    def get_cross_tags(crosscfg):
        return {
            'pyver': [crosscfg['implementation'] + crosscfg['version']],
            'abi': [crosscfg['abi']],
            'arch': [crosscfg['arch']],
        }

    def iter_files(self, stagedir):
        """Iterate over the files contained in the given folder.

        Yields absolute paths - caller may want to make them relative.
        Excludes any __pycache__ and *.pyc files."""

        # https://github.com/pypa/flit/blob/a4524758604107bde8c77b5816612edb76a604aa/flit_core/flit_core/common.py#L73

        def _include(path):
            name = os.path.basename(path)
            return name != '__pycache__' and not name.endswith('.pyc')

        # Ensure we sort all files and directories so the order is stable
        for dirpath, dirs, files in os.walk(str(stagedir)):
            for file in sorted(files):
                full_path = os.path.join(dirpath, file)
                if _include(full_path):
                    yield full_path

            dirs[:] = [d for d in sorted(dirs) if _include(d)]


_BACKEND = _BuildBackend()
get_requires_for_build_wheel = _BACKEND.get_requires_for_build_wheel
get_requires_for_build_sdist = _BACKEND.get_requires_for_build_sdist
get_requires_for_build_editable = _BACKEND.get_requires_for_build_editable
build_wheel = _BACKEND.build_wheel
build_sdist = _BACKEND.build_sdist
build_editable = _BACKEND.build_editable



"""
Classes to define hierarchical configuration options which support inheriting
from other options, default values, overriding options, etc.
"""

from abc import ABC, abstractmethod
from copy import deepcopy
from dataclasses import dataclass
import os
from pathlib import Path
from typing import Any, Dict, Iterable, Iterator, List, Optional, Tuple, Union

from flit_core.config import ConfigError

ConfPath = Tuple[str, ...]
ConfValue = Optional[Union[str, List[str], Dict[str, Any]]]
Conf = Dict[str, Any]


def pth(s: str) -> ConfPath:
    if not s:
        return ()
    return tuple(s.split('/'))


def pth2str(p: ConfPath) -> str:
    return '/'.join(p)


def joinpth(p1: ConfPath, p2: ConfPath) -> ConfPath:
    while p1 and p2 and p2[0] == '^':
        p1 = p1[:-1]
        p2 = p2[1:]
    return p1 + p2


def hasparent(path: ConfPath) -> bool:
    return len(path) >= 1


def parent(path: ConfPath) -> ConfPath:
    if not hasparent(path):
        raise RuntimeError(f"Path {pth2str(path)} does not have a parent")
    return path[:-1]


def basename(path: ConfPath) -> str:
    return path[-1]


class ConfigNode:

    def __init__(self,
                 value: ConfValue = None,
                 sub: Dict[str, 'ConfigNode'] = None) -> None:
        self.value: ConfValue = value
        self.sub: Optional[Dict[str, 'ConfigNode']] = sub

    @classmethod
    def from_dict(cls, d: dict):
        node = cls()
        node.sub = {}
        for k, v in d.items():
            if isinstance(v, dict):
                node.sub[k] = cls.from_dict(v)
            else:
                node.sub[k] = cls(value=v)
        return node

    def to_dict(self):
        if self.sub is None:
            return self.value
        return {k: v.to_dict() for k, v in self.sub.items()}

    def iter_dfs(self, path: ConfPath = ()):
        yield path, self.value
        if self.sub is not None:
            for name, sub in self.sub.items():
                for y in sub.iter_dfs(path + (name, )):
                    yield y

    def __getitem__(self, key):
        if isinstance(key, str):
            if self.sub is None:
                raise KeyError()
            return self.sub[key]
        elif isinstance(key, tuple):
            if len(key) == 0:
                return self
            elif self.sub is None:
                raise KeyError()
            else:
                return self.sub[key[0]][key[1:]]
        else:
            raise TypeError(key)

    def get(self, key: ConfPath, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def setdefault(self, path: ConfPath, default: Any):
        tgt = self[parent(path)]
        if tgt.sub is None:
            tgt.sub = {}
        return tgt.sub.setdefault(basename(path), default)

    def contains(self, path: ConfPath):
        try:
            self[path]
            return True
        except KeyError:
            return False


@dataclass
class DefaultValueWrapper:
    value: ConfValue


class DefaultValue(ABC):

    @abstractmethod
    def get_default(self, rootopts: 'ConfigOption', opt: 'ConfigOption',
                    cfg: ConfigNode, cfgpath: ConfPath,
                    optpath: ConfPath) -> Optional[DefaultValueWrapper]:
        ...

    @abstractmethod
    def get_name(self) -> str:
        ...


class DefaultValueValue(DefaultValue):

    def __init__(self, value: ConfValue) -> None:
        self.value: ConfValue = value

    def get_default(self, rootopts: 'ConfigOption', opt: 'ConfigOption',
                    cfg: ConfigNode, cfgpath: ConfPath,
                    optpath: ConfPath) -> Optional[DefaultValueWrapper]:
        return DefaultValueWrapper(self.value)

    def get_name(self):
        return repr(self.value)


class NoDefaultValue(DefaultValue):

    def get_default(self, rootopts: 'ConfigOption', opt: 'ConfigOption',
                    cfg: ConfigNode, cfgpath: ConfPath,
                    optpath: ConfPath) -> Optional[DefaultValueWrapper]:
        return None

    def get_name(self):
        return 'none'


class MissingDefaultError(ConfigError):
    pass


class RequiredValue(DefaultValue):

    def get_default(self, rootopts: 'ConfigOption', opt: 'ConfigOption',
                    cfg: ConfigNode, cfgpath: ConfPath,
                    optpath: ConfPath) -> Optional[DefaultValueWrapper]:
        raise MissingDefaultError(f'{pth2str(cfgpath)} requires a value')

    def get_name(self):
        return 'required'


class RefDefaultValue(DefaultValue):

    def __init__(self, path: ConfPath, relative: bool = False) -> None:
        super().__init__()
        self.path: ConfPath = path
        self.relative = relative

    def get_default(self, rootopts: 'ConfigOption', opt: 'ConfigOption',
                    cfg: ConfigNode, cfgpath: ConfPath,
                    optpath: ConfPath) -> Optional[DefaultValueWrapper]:
        abscfgpath = absoptpath = self.path
        if self.relative:
            absoptpath = joinpth(optpath, ('^', ) + absoptpath)
            abscfgpath = joinpth(cfgpath, ('^', ) + abscfgpath)
        opt = rootopts.get(absoptpath)
        if opt is None:
            raise ValueError("DefaultValue: reference to nonexisting option "
                             f"{pth2str(absoptpath)}")
        return opt.update_default(rootopts, cfg, abscfgpath, absoptpath)

    def get_name(self) -> str:
        r = pth2str(self.path).replace('^', '..')
        if not self.relative:
            r = '/' + r
        return r


class ConfigOption:

    allow_unknown_keys = False

    def __init__(self,
                 name: str,
                 description: str = '',
                 example: str = '',
                 default: DefaultValue = NoDefaultValue(),
                 inherit_from: Optional[ConfPath] = None,
                 create_if_inheritance_target_exists: bool = False) -> None:
        self.name = name
        self.description = description
        self.example = example
        self.sub: Dict[str, 'ConfigOption'] = {}
        self.default: DefaultValue = default
        self.inherit_from: Optional[ConfPath] = inherit_from
        self.create_if_inheritance_target_exists = create_if_inheritance_target_exists

    def get_typename(self):
        return None

    def insert(self, opt: 'ConfigOption'):
        assert opt.name not in self.sub
        self.sub[opt.name] = opt
        return self.sub[opt.name]

    def insert_multiple(self, opts: Iterable['ConfigOption']):
        for opt in opts:
            self.insert(opt)

    def iter_opt_paths(self) -> Iterator[ConfPath]:
        """DFS of the option tree."""
        for name, subopt in self.sub.items():
            yield (name, )
            for p in subopt.iter_opt_paths():
                yield (name, ) + p

    def iter_leaf_opt_paths(self) -> Iterator[ConfPath]:
        """DFS of the option tree."""
        if not self.sub:
            yield ()
        else:
            for name, subopt in self.sub.items():
                for p in subopt.iter_leaf_opt_paths():
                    yield (name, ) + p

    def iter_dfs(self, path: ConfPath = ()):
        yield path, self
        for name, sub in self.sub.items():
            for y in sub.iter_dfs(path + (name, )):
                yield y

    def __getitem__(self, key) -> 'ConfigOption':
        if isinstance(key, str):
            return self.sub[key]
        elif isinstance(key, tuple):
            if len(key) == 0:
                return self
            else:
                return self.sub[key[0]][key[1:]]
        else:
            raise TypeError(key)

    def get(self, key: ConfPath, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def setdefault(self, path: ConfPath, default: Any):
        tgt = self[parent(path)]
        if tgt.sub is None:
            tgt.sub = {}
        return tgt.sub.setdefault(basename(path), default)

    def contains(self, path: ConfPath):
        try:
            self[path]
            return True
        except KeyError:
            return False

    def inherit(self, rootopts: 'ConfigOption', cfg: ConfigNode,
                selfpth: ConfPath):
        superpth = self.inherit_from
        if superpth is not None:
            # If the super option is not set, there's nothing to inherit
            if (supercfg := cfg.get(superpth)) is None:
                return

            # If this option is not set, but the super option is,
            # create our own config as well, including all of its parents,
            # but only if create_if_inheritance_target_exists is set on those
            # options
            selfcfg = self.create_parent_config_for_inheritance(
                rootopts, cfg, selfpth)
            if selfcfg is None:
                return

            # Find the option we inherit from and make sure it exists
            if (superopt := rootopts.get(superpth)) is None:
                raise ValueError(f'{pth2str(superpth)} is not a valid option')

            # Create a copy of the config of our super-option and override it
            # with our own config
            supercfg = deepcopy(supercfg)
            superopt.explicit_override(self, supercfg, superpth, selfcfg,
                                       selfpth)
            selfcfg.sub = supercfg.sub
        if self.sub:
            for name, sub in self.sub.items():
                sub.inherit(rootopts, cfg, selfpth + (name, ))

    @staticmethod
    def create_parent_config_for_inheritance(rootopts: 'ConfigOption',
                                             cfg: ConfigNode,
                                             selfpth: ConfPath):
        """
        Loop over all parent options of selfpth in rootopts and default-
        initialize their configuration in cfg to empty ConfigNodes if the 
        option's create_if_inheritance_target_exists is set to True.
        Returns cfg[selfpth] or None if parents were not created because of
        create_if_inheritance_target_exists.
        """
        selfcfg = None
        p: ConfPath = ()
        opt = rootopts
        create_paths: List[ConfPath] = []
        for s in selfpth:
            p += s,
            opt = opt[s]
            if (selfcfg := cfg.get(p)) is None:
                if not opt.create_if_inheritance_target_exists:
                    return None
                create_paths.append(p)
        for p in create_paths:
            selfcfg = cfg.setdefault(p, ConfigNode(sub={}))
        return selfcfg

    def explicit_override(self, rootopts: 'ConfigOption', selfcfg: ConfigNode,
                          selfpth: ConfPath, overridecfg: ConfigNode,
                          overridepath: ConfPath):
        # The default ConfigOption simply overrides all of its sub-options, but
        # this function is overridden by specific subclasses.
        if overridecfg.sub is None:
            return  # No sub-options, so nothing to override
        for name, subopt in self.sub.items():
            assert isinstance(selfcfg, ConfigNode)
            assert isinstance(overridecfg, ConfigNode)
            if name not in overridecfg.sub:
                continue
            subselfcfg = selfcfg.setdefault((name, ), ConfigNode())
            subpath = selfpth + (name, )
            suboverridepath = overridepath + (name, )
            suboverridecfg = overridecfg.sub[name]
            subopt.explicit_override(rootopts, subselfcfg, subpath,
                                     suboverridecfg, suboverridepath)
        if self.inherit_from is not None:
            superopt = rootopts[self.inherit_from]
            superopt.explicit_override(rootopts, selfcfg, selfpth, overridecfg,
                                       overridepath)

    def override(self, rootopts: 'ConfigOption', cfg: ConfigNode,
                 selfpath: ConfPath):
        """Override other options with this option if appropriate. This is a 
        no-op in most cases and only does something in OverrideConfigOption."""
        assert cfg.contains(selfpath)

    def verify_impl(self, rootopts: 'ConfigOption', cfg: ConfigNode,
                    cfgpath: ConfPath):
        assert cfg.contains(cfgpath)
        selfcfg = cfg[cfgpath]
        # Check if there are any unknown options in the config
        if not self.allow_unknown_keys:
            if (unkwn := set(selfcfg.sub or ()) - set(self.sub or ())):
                raise ConfigError(f'Unkown options in {pth2str(cfgpath)}: ' +
                                  ', '.join(unkwn))
        # Recursively verify the sub-options
        if selfcfg.sub:
            for name, sub in selfcfg.sub.items():
                if name in self.sub:
                    self.sub[name].verify(rootopts, cfg, cfgpath + (name, ))

    def verify(self, rootopts: 'ConfigOption', cfg: ConfigNode,
               cfgpath: ConfPath):
        if self.inherit_from is None:
            return self.verify_impl(rootopts, cfg, cfgpath)
        else:
            return rootopts[self.inherit_from].verify_impl(
                rootopts, cfg, cfgpath)

    def override_all(self, cfg: ConfigNode):
        # This is just me being lazy, we probably don't need to iterate over
        # all nodes ...
        for p, opt in self.iter_dfs():
            if cfg.contains(p):
                opt.override(self, cfg, p)

    def verify_all(self, cfg: ConfigNode):
        self.verify(self, cfg, ())

    def inherit_all(self, cfg: ConfigNode):
        self.inherit(self, cfg, ())

    def update_default(
            self,
            rootopts: 'ConfigOption',
            cfg: ConfigNode,
            cfgpath: ConfPath,
            selfpath: Optional[ConfPath] = None
    ) -> Optional[DefaultValueWrapper]:
        if selfpath is None:
            selfpath = cfgpath

        result = None
        # If the entire path exists in cfg, simply return that value
        if (cfgval := cfg.get(cfgpath)) is not None:
            result = cfgval
        # If the path is not yet in cfg
        else:
            assert self is rootopts[selfpath]
            # Find the default value for this option
            default = self.default.get_default(rootopts, self, cfg, cfgpath,
                                               selfpath)
            # If the parent is set in the config, set this value as well
            if default is not None and cfg.contains(parent(cfgpath)):
                cfgval = cfg.setdefault(cfgpath, ConfigNode())
                cfgval.value = default.value
                if cfgval.value is not None:
                    self.verify(rootopts, cfg, cfgpath)
            result = default

        if self.inherit_from is not None:
            targetopt = rootopts.get(self.inherit_from)
            if targetopt is None:
                raise ValueError(f"Inheritance {pth2str(selfpath)} targets "
                                 f"nonexisting option "
                                 f"{pth2str(self.inherit_from)}")
            for p, opt in targetopt.iter_dfs():
                if opt.inherit_from is not None:
                    # TODO: this might be too restrictive, but we need to break
                    # the recursion somehow ...
                    continue
                optpth = joinpth(self.inherit_from, p)
                newcfgpth = joinpth(cfgpath, p)
                opt.update_default(rootopts, cfg, newcfgpth, optpth)

        return result

    def update_default_all(self, cfg: ConfigNode):
        for p, opt in self.iter_dfs():
            if hasparent(p) and cfg.contains(parent(p)):
                opt.update_default(self, cfg, p)


class UncheckedConfigOption(ConfigOption):

    allow_unknown_keys = True


class StrConfigOption(ConfigOption):

    def get_typename(self):
        return 'string'

    def explicit_override(self, opts: 'ConfigOption', selfcfg: ConfigNode,
                          selfpth: ConfPath, overridecfg: ConfigNode,
                          overridepath: ConfPath):
        assert not self.sub
        assert not selfcfg.sub
        assert not overridecfg.sub
        selfcfg.value = deepcopy(overridecfg.value)

    def verify(self, rootopts: 'ConfigOption', cfg: ConfigNode,
               cfgpath: ConfPath):
        if cfg[cfgpath].sub:
            raise ConfigError(f'Type of {pth2str(cfgpath)} should be '
                              f'{str}, not {dict}')
        elif not isinstance(cfg[cfgpath].value, str):
            raise ConfigError(f'Type of {pth2str(cfgpath)} should be '
                              f'{str}, not {type(cfg[cfgpath].value)}')


class PathConfigOption(StrConfigOption):

    def __init__(self,
                 name: str,
                 description: str = '',
                 example: str = '',
                 default: DefaultValue = NoDefaultValue(),
                 must_exist: bool = True,
                 expected_contents: List[str] = [],
                 base_path: Optional[Path] = None):
        super().__init__(name, description, example, default)
        self.must_exist = must_exist
        self.expected_contents = expected_contents
        self.base_path = base_path

    def get_typename(self):
        return 'path'

    def check_path(self, cfg: ConfigNode, cfgpath):
        path = cfg[cfgpath].value = os.path.normpath(cfg[cfgpath].value)
        if os.path.isabs(path):
            raise ConfigError(f'{pth2str(cfgpath)} must be a relative path')
        if self.base_path is not None:
            abspath = self.base_path / path
            if self.must_exist and not os.path.exists(abspath):
                raise ConfigError(f'{pth2str(cfgpath)}: {str(abspath)} '
                                  f'does not exist')
            missing = [
                sub for sub in self.expected_contents
                if not os.path.exists(os.path.join(abspath, sub))
            ]
            if missing:
                missingstr = '", "'.join(missing)
                raise ConfigError(f'{pth2str(cfgpath)} does not contain '
                                  f'required files or folders "{missingstr}"')

    def verify(self, rootopts: 'ConfigOption', cfg: ConfigNode,
               cfgpath: ConfPath):
        super().verify(rootopts, cfg, cfgpath)
        self.check_path(cfg, cfgpath)


class ListOfStrConfigOption(ConfigOption):

    def __init__(self,
                 name: str,
                 description: str = '',
                 example: str = '',
                 default: DefaultValue = NoDefaultValue(),
                 inherit_from: Optional[ConfPath] = None,
                 create_if_inheritance_target_exists: bool = False,
                 convert_str_to_singleton=False) -> None:
        super().__init__(name, description, example, default, inherit_from,
                         create_if_inheritance_target_exists)
        self.convert_str_to_singleton = convert_str_to_singleton

    def get_typename(self):
        return 'list'

    def explicit_override(self, opts: 'ConfigOption', selfcfg: ConfigNode,
                          selfpth: ConfPath, overridecfg: ConfigNode,
                          overridepath: ConfPath):
        assert not self.sub
        assert not selfcfg.sub
        assert not overridecfg.sub
        if overridecfg.value is not None:
            if selfcfg.value is None:
                selfcfg.value = []
            assert isinstance(selfcfg.value, list)
            assert isinstance(overridecfg.value, list)
            selfcfg.value += deepcopy(overridecfg.value)

    def verify(self, rootopts: 'ConfigOption', cfg: ConfigNode,
               cfgpath: ConfPath):
        if cfg[cfgpath].sub:
            raise ConfigError(f'Type of {pth2str(cfgpath)} should be '
                              f'{list}, not {dict}')
        elif not isinstance(cfg[cfgpath].value, list):
            if self.convert_str_to_singleton and \
                    isinstance(cfg[cfgpath].value, str):
                cfg[cfgpath].value = [cfg[cfgpath].value]
            else:
                raise ConfigError(f'Type of {pth2str(cfgpath)} should be '
                                  f'{list}, not {type(cfg[cfgpath].value)}')
        elif not all(isinstance(el, str) for el in cfg[cfgpath].value):
            raise ConfigError(f'Type of elements in {pth2str(cfgpath)} should '
                              f'be {str}')


class DictOfStrConfigOption(ConfigOption):

    def get_typename(self):
        return 'dict'

    def explicit_override(self, opts: 'ConfigOption', selfcfg: ConfigNode,
                          selfpth: ConfPath, overridecfg: ConfigNode,
                          overridepath: ConfPath):
        assert not self.sub
        assert not selfcfg.value
        assert not overridecfg.value
        if overridecfg.sub is not None:
            if selfcfg.sub is None:
                selfcfg.sub = {}
            assert isinstance(selfcfg.sub, dict)
            assert isinstance(overridecfg.sub, dict)
            selfcfg.sub.update(deepcopy(overridecfg.sub))

    def verify(self, rootopts: 'ConfigOption', cfg: ConfigNode,
               cfgpath: ConfPath):
        if cfg[cfgpath].value is not None:
            if isinstance(cfg[cfgpath].value, dict):
                newcfg = ConfigNode.from_dict(cfg[cfgpath].value)
                cfg[cfgpath].value = newcfg.value
                cfg[cfgpath].sub = newcfg.sub
            else:
                raise ConfigError(f'Type of {pth2str(cfgpath)} should be '
                                  f'{dict}, not {type(cfg[cfgpath].value)}')
        valdict = cfg[cfgpath].sub
        if not isinstance(valdict, dict):
            raise ConfigError(f'Type of {pth2str(cfgpath)} should be '
                              f'{dict}, not {type(valdict)}')
        elif not all(isinstance(el, str) for el in valdict.keys()):
            raise ConfigError(f'Type of keys in {pth2str(cfgpath)} should '
                              f'be {str}')
        elif not all(isinstance(el.value, str) for el in valdict.values()):
            raise ConfigError(f'Type of values in {pth2str(cfgpath)} should '
                              f'be {str}')


class OverrideConfigOption(ConfigOption):

    def __init__(
        self,
        name: str,
        description: str,
        targetpath: ConfPath,
        default: DefaultValue = NoDefaultValue()) -> None:
        super().__init__(name, description, '', default)
        self.targetpath = targetpath

    def verify(self, rootopts: 'ConfigOption', cfg: ConfigNode,
               cfgpath: ConfPath):
        rootopts[self.targetpath].verify(rootopts, cfg, cfgpath)

    def inherit(self, rootopts: 'ConfigOption', cfg: ConfigNode,
                selfpth: ConfPath):
        pass

    def override(self, rootopts: ConfigOption, cfg: ConfigNode,
                 cfgpath: ConfPath):
        if (selfcfg := cfg.get(cfgpath, None)) is None:
            return
        elif selfcfg.value is None and selfcfg.sub is None:
            return
        super().override(rootopts, cfg, cfgpath)
        curropt = rootopts[self.targetpath]
        self.create_parent_config(cfg, self.targetpath)
        currcfg = cfg[self.targetpath]
        overridecfg = cfg[cfgpath]
        # Override the config at those paths by our own config
        curropt.explicit_override(rootopts, currcfg, self.targetpath,
                                  overridecfg, cfgpath)

    @staticmethod
    def create_parent_config(cfg: ConfigNode, path: ConfPath):
        parentcfg = cfg
        for s in path:
            assert parentcfg.sub is not None
            parentcfg = parentcfg.sub.setdefault(s, ConfigNode(sub={}))



import html
import itertools
import shutil
import sys
import textwrap

from py_build_cmake.config_options import ConfigOption, NoDefaultValue, RefDefaultValue, RequiredValue, pth, pth2str
from py_build_cmake.pyproject_options import get_options


def _print_wrapped(text, indent, width=None):
    """Print the given string with the given indentation, wrapping it to the
    desired width, preserving line endings. If `width` is None, use the width 
    of the terminal, or 80 columns as a fallback."""
    if width is None:
        width = shutil.get_terminal_size((80, 24)).columns
    wrapper = textwrap.TextWrapper(width=width,
                                   initial_indent=indent,
                                   subsequent_indent=indent)
    wrapped = [wrapper.wrap(i) for i in text.split('\n')]
    for line in itertools.chain.from_iterable(wrapped):
        print(line)


def get_default_str(opt: ConfigOption):
    """Get a string representation of the default value for the given option."""
    return opt.default.get_name()


def help_print_md(pbc_opts: ConfigOption):
    """
    Prints the top-level options in `pbc_opts` as MarkDown tables.
    """
    for k, v in pbc_opts.sub.items():
        print('##', k)
        print(_get_full_description(v), '\n')
        print('| Option | Description | Type | Default |')
        print('|--------|-------------|------|---------|')
        for kk, vv in v.sub.items() or {}:
            print('|', f'`{kk}`', '|', _get_full_description(vv), '|',
                  vv.get_typename() or '', '|', f'`{get_default_str(vv)}`',
                  '|')
        print()


def _get_full_description(vv: ConfigOption):
    descr = _md_escape(vv.description)
    if vv.inherit_from:
        descr += '<br/>Inherits from: `/' + pth2str(vv.inherit_from) + '`'
    if vv.example:
        descr += '<br/>For example: `' + vv.example + '`'
    return descr


def _md_escape(descr):
    descr = html.escape(descr)
    descr = descr.replace('\n', '<br/>')
    descr = descr.replace('*', '\\*')
    descr = descr.replace('_', '\\_')
    return descr


def recursive_help_print(opt: ConfigOption, level=0):
    """Recursively prints the help messages for the options in `opt`."""
    for k, v in opt.sub.items():
        if k == 'project':
            continue
        indent = 4 * level * ' '
        header = '\n' + k
        if v.sub:
            header += ':'
            print(textwrap.indent(header, indent))
            if v.description:
                _print_wrapped(v.description, indent + '  ')
            recursive_help_print(v, level + 1)
        else:
            headerfields = []
            if (typename := v.get_typename()) is not None:
                headerfields += [typename]
            if is_required := isinstance(v.default, RequiredValue):
                headerfields += ['required']
            if v.inherit_from:
                headerfields += ['inherits from /' + pth2str(v.inherit_from)]
            if headerfields:
                header += ' (' + ', '.join(headerfields) + ')'
            print(textwrap.indent(header + ':', indent))
            _print_wrapped(v.description, indent + '  ')
            if v.example:
                _print_wrapped('For example: ' + v.example, indent + '  ')
            default = v.default.get_name()
            if default is not None and not is_required:
                print(textwrap.indent('Default: ' + default, indent + '  '))


def _print_usage():
    print(
        textwrap.dedent("""\
    Print the documentation for the pyproject.toml configuration options for the
    py-build-cmake build backend.

    Usage:
    """))
    name = __loader__.name.rsplit(sep='.', maxsplit=1)[0]
    print("   ", sys.executable, "-m", name, "[md]")
    print(
        textwrap.dedent("""
    Use the 'md' option to print the options as a MarkDown table instead of a
    plain-text list.
    """))


if __name__ == '__main__':
    opts = get_options()
    help_pth = pth('pyproject.toml/tool/py-build-cmake')
    help_opt = {'-h', '-?', '--help', 'h', 'help', '?'}
    if len(sys.argv) == 2 and sys.argv[1] == 'md':
        pbc_opts = opts[help_pth]
        print("# py-build-cmake configuration options\n")
        print("These options go in the `[tool.py-build-cmake]` section of "
              "the `pyproject.toml` configuration file.\n")
        help_print_md(pbc_opts)
        print("# Local overrides\n")
        print("Additionally, two extra configuration files can be placed in "
              "the same directory as `pyproject.toml` to override some "
              "options for your specific use case:\n\n"
              "- `py-build-cmake.local.toml`: the options in this file "
              "override the values in the `tool.py-build-cmake` section of "
              "`pyproject.toml`.<br/>This is useful if you need specific "
              "arguments or CMake options to compile the package on your "
              "system.\n"
              "- `py-build-cmake.cross.toml`: the options in this file "
              "override the values in the `tool.py-build-cmake.cross` section "
              "of `pyproject.toml`.<br/>Useful for cross-compiling the "
              "package without having to edit the main configuration file.\n")
    elif len(sys.argv) > 1 or set(map(str.lower, sys.argv[1:])) & help_opt:
        _print_usage()
    else:
        print("List of py-build-cmake pyproject.toml options:")
        recursive_help_print(opts[help_pth])
        print()
